# TODO check these lists resources
# List of exclusive SYSDAMIN role policies
# The role SYSADMIN is exclusive for DEFAULT application
SYSADMIN_EXCLUSIVE_POLICIES = [
    ('/applications', ['POST']),
    ('/applications/<id>', ['PUT', 'DELETE']),

    ('/domains', ['POST', 'GET']),
    ('/domains/<id>', ['DELETE']),

    ('/capabilities', ['POST']),
    ('/capabilities/<id>', ['PUT', 'DELETE']),

    ('/policies', ['POST']),
    ('/policies/<id>', ['PUT', 'DELETE']),

    ('/roles', ['POST']),
    ('/roles/<id>', ['PUT', 'DELETE']),
    ('/roles/<id>/policies', ['POST']),

    ('/routes', ['POST']),
    ('/routes/<id>', ['PUT', 'DELETE'])
]

SYSADMIN_RESOURCES = [
    ('/applications', ['GET']),
    ('/applications/<id>', ['GET']),
    ('/applications/<id>/roles', ['GET']),

    ('/domains/<id>', ['PUT', 'GET']),
    ('/domains/<id>/settings', ['POST', 'PUT', 'DELETE']),
    ('/domains/<id>/logo', ['PUT', 'DELETE']),

    ('/capabilities', ['GET']),
    ('/capabilities/<id>', ['GET']),

    ('/policies', ['GET']),
    ('/policies/<id>', ['GET']),

    ('/roles', ['GET']),
    ('/roles/<id>', ['GET']),

    ('/images', ['POST', 'GET']),
    ('/images/<id>', ['PUT', 'DELETE']),

    ('/files', ['POST', 'GET']),
    ('/files/<id>', ['PUT', 'GET', 'DELETE']),

    ('/routes', ['POST', 'GET']),
    ('/routes/<id>', ['PUT', 'GET', 'DELETE']),

    ('/grants', ['POST', 'GET']),
    ('/grants/<id>', ['PUT', 'GET', 'DELETE']),

    ('/notifications', ['POST', 'GET']),
    ('/notifications/<id>', ['PUT', 'GET', 'DELETE']),

    ('/tags', ['POST', 'GET']),
    ('/tags/<id>', ['PUT', 'GET', 'DELETE']),

    ('/tokens', ['GET']),
    ('/tokens/<id>', ['DELETE']),

    ('/users', ['POST', 'GET']),
    ('/users/<id>', ['DELETE']),
    ('/users/<id>/reset', ['POST'])
]

# Common resources for all users
USER_RESOURCES = [
    ('/tokens', ['POST']),
    ('/tokens/<id>', ['GET', 'DELETE']),
    ('/applications/<id>', ['GET']),
    ('/domains/<id>', ['GET']),
    ('/domains/<id>/settings', ['GET']),
    ('/domain_sequences/<id>/nextval', ['PUT']),
    ('/images/<id>', ['GET']),

    ('/users/<id>', ['GET', 'PUT']),
    ('/users/<id>/photo', ['PUT', 'DELETE']),
    ('/users/<id>/notify', ['POST']),
    ('/users/<id>/update_my_password', ['PUT']),
    ('/users/routes', ['GET']),
    ('/users/reset', ['POST']),
    ('/timeline_events', ['GET'])
]



from viggocore.common import subsystem
from viggocore.subsystem.capability import resource, manager


subsystem = subsystem.Subsystem(resource=resource.Capability,
                                manager=manager.Manager)



from typing import List
import flask

from viggocore.common import exception, utils
from viggocore.common.input import InputResource
from viggocore.common.subsystem import controller


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super().__init__(manager, resource_wrap, collection_wrap)

    def _get_resources(self, exceptions) -> List[InputResource]:
        resources = []
        for exception_resource in exceptions:
            endpoint = exception_resource.get('endpoint')
            methods = exception_resource.get('methods', [])
            resource = (endpoint, methods)
            resources.append(resource)

        return resources

    def create(self):
        data = flask.request.get_json()

        try:
            if data:
                exceptions = data.get('exceptions', None)
                if exceptions:
                    data['exceptions'] = self._get_resources(exceptions)
                entity = self.manager.create(**data)
            else:
                entity = self.manager.create()
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        response = {self.resource_wrap: entity.to_dict()}

        return flask.Response(response=utils.to_json(response),
                              status=201,
                              mimetype="application/json")

    def get_roles(self, id: str):
        try:
            roles = self.manager.get_roles(id=id)
            response = {"roles": [role.to_dict() for role in roles]}

        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")

    def update_settings(self, id):
        try:
            data = flask.request.get_json()

            settings = self.manager.update_settings(id=id, **data)
            response = {'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def _get_keys_from_args(self):
        keys = flask.request.args.get('keys')
        if not keys:
            raise exception.BadRequest(
                'ERRO! The keys parameter was not passed correctly')
        return list(filter(None, keys.split(',')))

    def remove_settings(self, id):
        try:
            keys = self._get_keys_from_args()
            kwargs = {'keys': keys}

            settings = self.manager.remove_settings(id=id, **kwargs)
            response = {'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

    def get_application_settings_by_keys(self, id):
        try:
            keys = self._get_keys_from_args()
            kwargs = {'keys': keys}

            settings = self.manager.get_application_settings_by_keys(
                id=id, **kwargs)
            response = {'id': id, 'settings': settings}

            return flask.Response(response=utils.to_json(response),
                                  status=200,
                                  mimetype="application/json")
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)



import uuid
import json
from typing import Any
from sqlalchemy import orm
from viggocore.database import db
from sqlalchemy import UniqueConstraint
from viggocore.common import exception
from viggocore.common.subsystem import entity


class User(entity.Entity, db.Model):

    SYSADMIN_USERNAME = 'sysadmin'

    attributes = ['domain_id', 'name', 'email', 'nickname', 'photo_id',
                  'settings']
    attributes += entity.Entity.attributes

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain.id'), nullable=False)
    domain = orm.relationship('Domain', backref=orm.backref('users'))
    name = db.Column(db.String(80), nullable=False)
    nickname = db.Column(db.String(80), nullable=True)
    email = db.Column(db.String(80), nullable=False)
    password = db.Column(
        db.String(64), nullable=False, default=uuid.uuid4().hex)
    photo_id = db.Column(db.CHAR(32), db.ForeignKey('image.id'), nullable=True)
    _settings = db.Column('settings', db.Text, nullable=False, default='{}')

    tokens = orm.relationship('Token', backref=orm.backref('user_tokens'))

    __table_args__ = (
        UniqueConstraint('name', 'domain_id', name='user_name_domain_id_uk'),
        UniqueConstraint(
            'email', 'domain_id', name='user_email_domain_id_uk'),)

    def __init__(self, id, domain_id, name, email,
                 nickname=None, photo_id=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.name = name
        self.email = email
        # self.password = password
        self.nickname = nickname
        self.photo_id = photo_id

    def _has_setting(self, key: str) -> bool:
        return self.settings.get(key) is not None

    def remove_setting(self, key: str):
        if not self._has_setting(key):
            raise exception.BadRequest(f"Erro! Setting {key} not exists")

        settings = self.settings
        value = settings.pop(key)
        self._save_settings(settings)

        return value

    def update_setting(self, key: str, value: Any):
        settings = self.settings
        settings[key] = value
        self._save_settings(settings)
        return value

    @property
    def settings(self):
        try:
            settings_str = '{}' if self._settings is None else self._settings
            return json.loads(settings_str)
        except Exception:
            return {}

    def _save_settings(self, settings: dict):
        self._settings = json.dumps(settings, default=str)



import json

from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Optional, Union, Any
import uuid

DATE_FMT = '%Y-%m-%d'
DATETIME_FMT = '%Y-%m-%dT%H:%M:%S.%fZ'


# ----------
# Date functions
# ----------
def convert_date_from_str(value: Optional[str]) -> Optional[datetime]:
    date_time = None
    if value is not None:
        try:
            if len(value.strip()) == 10:
                date_time = datetime.strptime(value, DATE_FMT)
            elif len(value.strip()) == 24:
                date_time = datetime.strptime(value, DATETIME_FMT)
        except Exception:
            pass
    return date_time


def convert_date_to_str(date_value: Union[date, datetime]) -> str:
    if type(date_value) is date:
        return date_value.strftime(DATE_FMT)
    elif type(date_value) is datetime:
        return date_value.strftime(DATETIME_FMT)
    else:
        raise Exception('{} is not date or datetime'.format(date_value))
# ----------


# ----------
# Decimal functions
# ----------
def to_decimal(num):
    try:
        return Decimal(str(num))
    except Exception:
        return Decimal('0.0')


def to_decimal_n(num, n):
    try:
        num_str = '{:.' + str(n) + 'f}'
        num_str = num_str.format(num)
        return Decimal(num_str)
    except Exception:
        return Decimal('0.0')


def is_decimal(num):
    response = False
    if (to_decimal(num) - int(num)) > to_decimal('0.0'):
        response = True
    return response


def convert_decimal(value: Decimal) -> Union[float, int]:
    if isinstance(value, Decimal):
        if not value.is_finite():
            return str(value)
        if str(value).find('.') > -1:
            return float(value.real)
        else:
            return int(value)
    else:
        raise Exception('{} is not Decimal'.format(value))
# ----------


class ViggocoreEncoder(json.JSONEncoder):

    def default(self, value: Any) -> Any:
        try:
            if isinstance(value, Decimal):
                return convert_decimal(value)
            if isinstance(value, date):
                return convert_date_to_str(value)
            if isinstance(value, Enum):
                return value.name

            # default  str
            return str(value)
        except Exception:
            return super().default(value)


def to_json(value: Any) -> str:
    return json.dumps(value, cls=ViggocoreEncoder)


def random_uuid() -> str:
    return uuid.uuid4().hex



from celery.app.utils import Settings
from viggocore.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        settings_endpoint = '/settings'
        return super().routes + [
            {
                'action': 'Get Domain By Name',
                'method': 'GET',
                'url': '/domainbyname',
                'callback': 'domain_by_name',
                'bypass': True
            },
            {
                'action': 'Get Domain Logo By Name',
                'method': 'GET',
                'url': '/domainlogobyname',
                'callback': 'domain_logo_by_name',
                'bypass': True
            },
            {
                'action': 'Upload logo to Domain',
                'method': 'PUT',
                'url': self.resource_url + '/logo',
                'callback': 'upload_logo'
            },
            {
                'action': 'Remove logo from Domain',
                'method': 'DELETE',
                'url': self.resource_url + '/logo',
                'callback': 'remove_logo'
            },
            {
                'action': 'Register new Domain',
                'method': 'POST',
                'url': self.collection_url + '/register',
                'callback': 'register',
                'bypass': True
            },
            {
                'action': 'Activate a register Domain',
                'method': 'PUT',
                'url': self.resource_enum_url + '/activate/<id2>',
                'callback': 'activate',
                'bypass': True
            },
            {
                'action': 'Update settings on Domain',
                'method': 'PUT',
                'url': self.resource_url + settings_endpoint,
                'callback': 'update_settings',
                'bypass': False
            },
            {
                'action': 'Remove settings from Domain',
                'method': 'DELETE',
                'url': self.resource_url + settings_endpoint,
                'callback': 'remove_settings',
                'bypass': False
            },
            {
                'action': 'Get settings by keys from Domain',
                'method': 'GET',
                'url': self.resource_url + settings_endpoint,
                'callback': 'get_domain_settings_by_keys',
                'bypass': False
            },
            {
                'action': 'Send email activate account',
                'method': 'POST',
                'url': self.collection_url + '/send_email_activate_account',
                'callback': 'send_email_activate_account',
                'bypass': True
            }
        ]



import uuid
import hashlib

from viggocore.common import exception
from viggocore.common.subsystem import manager
from viggocore.common.subsystem import operation


class Create(operation.Operation):

    def pre(self, **kwargs):
        # FIXME(samueldmq): this method needs to receive the parameters
        # explicitly.
        if kwargs.get('user'):
            # FIXME(samueldmq): how to avoid someone simply passing the user
            # in the body and then having a valid token?
            self.user = kwargs['user']
        else:
            domain_name = kwargs.get('domain_name', None)
            username = kwargs.get('username', None)
            email = kwargs.get('email', None)
            password = kwargs.get('password', None)
            password_hash = kwargs.get('password_hash', None)

            # TODO(samueldmq): allow get by unique attrs
            domains = self.manager.api.domains().list(name=domain_name)

            if not domains:
                return False

            domain_id = domains[0].id
            if password_hash is None:
                password_hash = hashlib.sha256(
                    password.encode('utf-8')).hexdigest()

            if (email is None):
                users = self.manager.api.users().list(
                    domain_id=domain_id, name=username, password=password_hash)
            else:
                users = self.manager.api.users().list(
                    domain_id=domain_id, email=email, password=password_hash)

            if not users:
                return False

            self.user = users[0]
            if self.user.active is False:
                raise exception.PreconditionFailed('Usuário não está ativo!')

        return self.user.is_stable()

    def do(self, session, **kwargs):
        # TODO(samueldmq): use self.user.id instead of self.user_id
        token = self.driver.instantiate(
            id=uuid.uuid4().hex,
            created_by=self.user.id,
            user_id=self.user.id)

        self.driver.create(token, session=session)

        return token


class Manager(manager.Manager):

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)



import flask

from pika import BlockingConnection, PlainCredentials, \
    ConnectionParameters, BasicProperties
from pika.exchange_type import ExchangeType


class RabbitMQ:

    def __init__(self):
        self.url = flask.current_app.config['VIGGOCORE_QUEUE_URL']
        self.port = flask.current_app.config['VIGGOCORE_QUEUE_PORT']
        self.virtual_host = \
            flask.current_app.config['VIGGOCORE_QUEUE_VIRTUAL_HOST']
        self.username = flask.current_app.config['VIGGOCORE_QUEUE_USERNAME']
        self.password = flask.current_app.config['VIGGOCORE_QUEUE_PASSWORD']
        credentials = PlainCredentials(self.username, self.password)
        self.params = ConnectionParameters(
            self.url, self.port, self.virtual_host, credentials, heartbeat=60,
            blocked_connection_timeout=30, socket_timeout=None, retry_delay=10,
            connection_attempts=10000)

    def connect(self):
        try:
            return BlockingConnection(self.params)
        except Exception:
            raise


class BasicQueueConsumer():

    queue_name = None
    arguments = None
    exchanges = []

    def __init__(self, queue_name, exchanges=None, prefetch_size=None,
                 arguments=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        rabbitmq = RabbitMQ()
        self.connection = rabbitmq.connect()
        self.channel = self.connection.channel()
        if prefetch_size is not None:
            self.channel.basic_qos(prefetch_count=prefetch_size)
        self.queue_name = queue_name
        self.arguments = arguments
        self.exchanges = exchanges

    def __bind_queue(self):
        for ex in self.exchanges:
            exchange = ex.get('exchange', '')
            rounting_key = ex.get('routing_key', '')
            ex_type = ex.get('type', 'direct')
            self.channel.exchange_declare(exchange=exchange,
                                          exchange_type=ExchangeType[ex_type],
                                          durable=True)

            self.channel.queue_bind(self.queue_name, exchange=exchange,
                                    routing_key=rounting_key)

    def declare_and_consume(self, callback):
        self.channel.queue_declare(queue=self.queue_name, durable=True,
                                   arguments=self.arguments)
        self.channel.basic_consume(queue=self.queue_name, auto_ack=False,
                                   on_message_callback=callback)
        self.__bind_queue()

        self.channel.start_consuming()

    def close(self):
        self.channel.stop_consuming()
        self.channel.close()
        self.connection.close()

    def count_rejects_this_message(self, properties):
        count = 0
        xdeath = properties.headers.get('x-death', [])
        if len(xdeath) > 0:
            count = xdeath[0].get('count', 0)
        return count


class ProducerQueue:

    def __init__(self):
        rabbitmq = RabbitMQ()
        self.connection = rabbitmq.connect()
        self.channel = self.connection.channel()

    def publish(self, exchange, routing_key, body, properties=None):
        self.channel.basic_publish(exchange=exchange,
                                   routing_key=routing_key,
                                   body=body,
                                   properties=properties)

    def _publish_entity(self, exchange, routing_key, body,
                        type, priority=None, headers=None):
        properties = BasicProperties(
            type=type, headers=headers, priority=priority)
        self.channel.basic_publish(exchange=exchange,
                                   routing_key=routing_key,
                                   body=body,
                                   properties=properties)

    def publish_full_entity(self, exchange, routing_key, body,
                            type, priority):
        headers = {'event_type': 'FULL_ENTITY'}
        self._publish_entity(exchange, routing_key, body,
                             type, priority, headers)

    def publish_request_entity(self, exchange, routing_key, body,
                               type, priority):
        headers = {'event_type': 'REQUEST_ENTITY'}
        self._publish_entity(exchange, routing_key, body,
                             type, priority, headers)

    def publish_partial_entity(self, exchange, routing_key, body,
                               type, priority, event_name):
        headers = {'event_type': 'PARTIAL_ENTITY', 'event_name': event_name}
        self._publish_entity(exchange, routing_key, body,
                             type, priority, headers)

    def run(self, fn, *args):
        fn(self, *args)
        self.close()

    def close(self):
        self.channel.close()
        self.connection.close()



import flask

from viggocore.common import utils
from viggocore.common.subsystem import controller
from viggocore.common import exception


class Controller(controller.Controller):

    def __init__(self, manager, resource_wrap, collection_wrap):
        super(Controller, self).__init__(
            manager, resource_wrap, collection_wrap)

    def get_tags_from_entity(self):
        filters = self._filters_parse()
        filters = self._filters_cleanup(filters)

        try:
            filters = self._parse_list_options(filters)
            tags = self.manager.get_tags_from_entity(**filters)
        except exception.ViggoCoreException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)
        except ValueError:
            raise exception.BadRequest('page or page_size is invalid')

        collection = tags

        response = {self.collection_wrap: collection}

        return flask.Response(response=utils.to_json(response),
                              status=200,
                              mimetype="application/json")
